#ifndef __RAW2NEF_H
#define __RAW2NEF_H
//
#define PROGRAM_VERSION "raw2nef v0.14"

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;
typedef unsigned short color_t;


class TBuff
{
public:
   TBuff() : p(NULL) {}
   TBuff(unsigned sz) : p(new color_t[sz]) {}
   ~TBuff() { delete [] p; }

   color_t* p;
   operator color_t* () { return p; }
   color_t& operator [] (int i) { return p[i]; }

   bool alloc(unsigned sz) { delete [] p; return NULL != (p = new color_t[sz]); }
   void free() { delete [] p; p = NULL; }
};

struct unpack_buff { char buf[12]; };
const char* unpack( unpack_buff& buf, unsigned x );
unsigned pack( const char* s );

#define CAMERA_ID(id) pack(#id)

struct TCCDParam
{
   unsigned raw_width;
   unsigned raw_height;
   unsigned file_size;
   unsigned cfa_colors;
   unsigned row_length;
   unsigned bits_per_sample;
   unsigned camera;
   unsigned data_offset;

   enum
   {
      GMYC, MGCY, YCGM, CYMG,
      BGGR, GBRG, GRBG, RGGB,
      ShiftRightBit = 1,
      ShiftDownBit = 2,
      PrimaryColorsBit = 4,
   };
   enum
   {
      fInterlaced       = 0x00000001,
      fARM              = 0x00000002,
      fIntelByteOrder   = 0x00000004,
   };

   unsigned flags;
};

enum EReadResult
{
   e_success,
   e_unexpected_eof, // incomplete file ?
   e_too_large_file, // bad format description ?
   e_not_supported, // format not supported
};

int read_raw(FILE* in, color_t ccd[], TCCDParam& ccd_pars);


#ifndef __TIFF_H
#include "tiff.h"
#endif

class NEF_IFD
{
public:
   IFDir IFD0;
   IFDir EXIF;
   IFDir MakerNote;
   IFDir IFD;

   bool read_exif(FILE*);
   bool write_nef(FILE*, bool, color_t[], TCCDParam&, bool=false);
};

struct TRaw2Nef
{
   TRaw2Nef();
   ~TRaw2Nef();

   char* ini_fn; // ini-file name
   char* ifn; // input RAW file name
   char* ofn; // output NEF file name
   char* efn; // EXIF file name

   char* exe_fn; // this program path and filename

   char* ofn_buf;
   char* efn_buf;
   char* ini_buf;

   bool verbose;
   bool copy_exif; // copy EXIF metadata from paired JPEG to NEF
   bool endian;   // endian: true for Intel, false for Motorola
   bool compatible;
   bool incomplete; // allows load incomplete raw files
   bool same_number; // EXIF file has the same number as RAW file

   bool ProcessFile(unsigned camera);
   bool ParseIniFile(TCCDParam& ccd_pars, unsigned camera, unsigned file_size);
   void CreateExifFileName(const char* FilePath, int ExtOffset);

   bool raw_to_nef(int n, char* a[]);
};

#endif
